%
% OKECA Compute the principal components of OKECA method.
%
% [AA MMM V D M] = okeca(K,GD_steps,n_inizial,tol_tau)
% 
% inputs:
% 
% - K:          Kernel matrix 
% - GD_steps:   Number of gradient descent steps (default = 1000)
% - n_inizial:  Number of random inicializations (default = 50)
% - tol_tau:    Error tolerance (default = 2*eps)
% 
% outputs:
% 
% - AA:         OKECA features in colums, K = AA*AA'; 
% - MMM:        Values of the cost function
% - V:          Eigenvectors of K
% - D:          Eigenvalues of K
% - M:          Extra rotation matrix, i.e AA = V*(D.^0.5)*M
%
% Copyright (c) 2013  Emma Izquierdo-Verdiguier, Valero Laparra, 
%                     Robert Jenssen, Luis Gómez-Chova, and Gustavo Camps-Valls
%
%                     emma.izquierdo@uv.es, http://isp.uv.es
%

function [AA MMM V D M] = okeca(K,GD_steps,n_inizial,tol_tau)

DIM = size(K,1);

if ~exist('GD_steps','var')
    GD_steps = 1000;
end

if ~exist('n_inizial','var')
    n_inizial = 50;
end

if ~exist('tol_tau','var')
    tol_tau = 2*eps;
end


% Eigendecomposition

M = zeros(DIM);
D = -5;
kk = 0;
while(min(diag(D))<0 || sum(abs(imag(diag(D))))>0)
    [V D] = eigs(full(K + kk*eye(size(K,1))),DIM);
    kk = kk + max([sum(abs(imag(diag(D)))) abs(min(diag(D)))]);
end

[ddd iii] = sort(diag(D));
V = V(:,iii);
dD = diag(D);
D = diag(dD(iii));
Q = V*(D.^0.5);

RES = [];

aaa = sum(Q).^2;
[sa ia] = sort(aaa,'descend');

for di = 1:DIM
    
    % Optimizing for different initializations
    
    RES(n_inizial).End = 0;
    RES(n_inizial).m = zeros(DIM,1);
    
    for real = 1:n_inizial
        
        % Vector initialization
        % the first realization departs from the classical KECA solution
        
        if real == 1
            m = zeros(DIM,1);
            m(ia(di)) = 1;
            m = m - sum((repmat((M'*m)',DIM,1).*M),2);
            m = m/sqrt(sum(m.^2));
        else
            m = rand(DIM,1);
            m = m - sum((repmat((M'*m)',DIM,1).*M),2);
            m = m/sqrt(sum(m.^2));
        end
        
        if di<DIM
            tau = 1;
            E = zeros(1,GD_steps);
            E(1) = sum(Q*m).^2;
            
            MS(GD_steps).m = zeros(size(m));
            MS(1).m = m;
            
            for it = 1:GD_steps
                
                % Gradient descent step
                dJ = 2*(sum(Q*m))*sum(Q,1)';
                mn = m + tau*dJ;
                
                mn = mn - sum((repmat((M'*mn)',DIM,1).*M),2);
                mn = mn/sqrt(sum(mn.^2));
                
                E(it+1) = sum(Q*mn).^2;
                
                if it>0
                    while (E(it+1) <= E(it)) && abs(tau)>tol_tau
                        tau = -tau*0.5;
                        mn = m + tau*dJ;
                        
                        mn = mn - sum((repmat((M'*mn)',DIM,1).*M),2);
                        mn = mn/sqrt(sum(mn.^2));
                        
                        E(it+1) = sum(Q*mn).^2;
                    end
                end
               
                if abs(tau)<=tol_tau
                    break
                end
                
                m = mn;
                MS(it+1).m = m;
            end
            
            [mm ii] = max(E);
            RES(real).m = MS(ii).m;
            RES(real).End = E(end-1);
        else
            RES(real).m = m;
            RES(real).End = 1;
        end
        
    end
    
    % Selection of the best initialization
    [mmm iii] = max(cat(1, RES.End));
    
    MMM(di) = mmm;
    
    M(:,di) = RES(iii).m;
end

AA = Q*M;

% Sometimes Gradient descent may get stuck into local minima. Resorting the vectors.

valores_keca_2 = sum(AA).^2;
[ss iii] = sort(valores_keca_2,'descend');
AA = AA(:,iii);


MMM = MMM(iii);
M = M(:,iii);


